from collections.abc import Iterator, Sized
from typing import Any, override

import torch
from torch.utils.data import Dataset, Sampler


class RandomTimeSeriesSampler(Sampler[list[int]]):
    """A sampler that randomly samples continuous sequences from a time series
    dataset.

    Each sample contains a list of consecutive indices of specified
    length, starting from a randomly selected position in the entire
    dataset.
    """

    def __init__(
        self,
        dataset: Dataset[Any],
        sequence_length: int,
        max_samples: int | None = None,
    ) -> None:
        """
        Args:
            dataset (Dataset): The dataset to sample from
            sequence_length (int): Length of each sequence.
            max_samples (int | None): Maximum number of samples to generate
        """

        if not isinstance(dataset, Sized):
            raise TypeError("dataset must implement __len__")
        if len(dataset) < sequence_length:
            raise ValueError(
                "dataset length must be greater than or equal to sequence_length"
            )

        self.dataset = dataset
        self.sequence_length = sequence_length
        self.max_samples = (
            len(self.dataset) - self.sequence_length + 1
            if max_samples is None
            else max_samples
        )

    def __len__(self) -> int:
        """Returns the total number of batches generated by the sampler."""
        return min(len(self.dataset) - self.sequence_length + 1, self.max_samples)

    @override
    def __iter__(self) -> Iterator[list[int]]:
        """Returns an iterator producing sequences of consecutive indices.

        Each sequence starts at a random position in the dataset and has
        a length of sequence_length.
        """
        start_indices = torch.randperm(len(self.dataset) - self.sequence_length + 1)[
            : len(self)
        ].tolist()
        for start in start_indices:
            yield torch.arange(start, start + self.sequence_length).tolist()
